@@ -9,19 +9,19 @@ repository-independent way.
 #
 # This file is part of cfvers.
 #
-# cfvers is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
+# cfvers is free software; you can redistribute it and/or modify it
+# under the terms of the GNU General Public License as published by
 # the Free Software Foundation; either version 2 of the License, or
 # (at your option) any later version.
 #
-# cfvers is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU General Public License for more details.
+# cfvers is distributed in the hope that it will be useful, but
+# WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+# General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with cfvers; if not, write to the Free Software
-# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+# along with cfvers; if not, write to the Free Software Foundation,
+# Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 
 import os, struct, stat, os.path, re, commands, sys
 import base64
@@ -30,10 +30,13 @@ import random
 import difflib
 import time
 import bz2
+import sha
+import errno
+import string
 from cStringIO import StringIO
 from mx import DateTime
 
-__all__ = ["Area", "Item", "RevEntry", "AreaRev", "forcejoin"]
+__all__ = ["Area", "Item", "RevEntry", "AreaRev", "forcejoin", "rexists"]
 
 def forcejoin(a, *p):
     """Join two or more pathname components, considering them all relative"""
@@ -46,6 +49,15 @@ def forcejoin(a, *p):
         else:
             path = path + b
     return path
+
+def rexists(filename):
+    """Checks if a file exists, even if is a broken symlink"""
+    try:
+        os.lstat(filename)
+    except OSError, e:
+        if e.errno == errno.ENOENT:
+            return False
+    return True
             
 class Area(object):
     """Class implementing an area object.
@@ -56,27 +68,25 @@ class Area(object):
     jails on one server.
     
     """
-    __slots__ = ["server", "id", "revno", "_ctime", "name", "root", "description", "numitems"]
+    __slots__ = ["name", "revno", "_ctime", "root",
+                 "description", "numitems", "revno"]
     
-    def __init__(self, id=None, server=None, name="default",
-                 description="", root="/", revno=0, ctime = None,
-                 numitems=0):
+    def __init__(self, name=None,
+                 description="", root="/", ctime = None,
+                 numitems=0, revno=0):
         """Constructor for the Area class"""
-        if server is None:
-            self.server = os.uname()[1]
-        else:
-            self.server = server
-        self.id = id
-        self.revno = revno
-        self.ctime = ctime or DateTime.utc()
         self.name = name
+        self.ctime = ctime or DateTime.utc()
         self.description = description
         self.root = root
-        self.numitems = numitems
+        self.numitems = int(numitems)
+        self.revno = revno
+        return
 
     def _set_ctime(self, val):
         if isinstance(val, types.StringTypes):
             val = DateTime.ISO.ParseDateTimeUTC(val)
+            val = val.localtime()
         self._ctime = val
 
     def _get_ctime(self):
@@ -85,33 +95,21 @@ class Area(object):
     ctime = property(_get_ctime, _set_ctime, None,
                      "The creation time of this area")
     
-    def parse(repo, value):
-        """Converts a string representation to an server, area object pair"""
-        try:
-            areaid = int(value)
-        except ValueError:
-            areaid = None
-        if areaid is not None:
-            return repo.getArea(areaid)
-        m = re.match("^((?P<server>.*?):)?(?P<area>.*)$", value)
-        if m is None:
-            raise ValueError, "Invalid area specification: %s" % value
-        server = m.group('server')
-        if server is None:
-            server = os.uname()[1]
-        return repo.getAreaByName(server, m.group('area'))
-    parse = staticmethod(parse)
-
 
 class Item(object):
-    __slots__ = ["id", "area", "name", "ctime"]
+    __slots__ = ["id", "area", "name", "ctime", "dirname"]
     
-    def __init__(self, id=-1, area=None, name=None, ctime=None):
+    def __init__(self, id=-1, area=None, name=None,
+                 ctime=None, dirname=None):
         if name is None or not name.startswith("/"):
             raise ValueError, "Invalid name '%s'" % name
         self.id = id
         self.area = area
         self.name = name
+        if dirname is None:
+            self.dirname = os.path.dirname(name)
+        else:
+            self.dirname = dirname
         if ctime is None:
             self.ctime = DateTime.utc()
         elif not isinstance(ctime, DateTime.DateTimeType):
@@ -124,13 +122,31 @@ class Item(object):
         return "Item: id #%d, name %s" % (self.id, self.name)
 
 class RevEntry(object):
-    __slots__ = ["item", "revno", "filename", "filetype", "filecontents",
-                 "mode", "mtime", "atime", "uid", "gid", "rdev",
-                 "encoding",]
+    __slots__ = ["item", "revno",
+                 "filename", "filetype",
+                 "filecontents",
+                 "mode",
+                 "mtime", "atime", "ctime",
+                 "inode", "device", "nlink",
+                 "uid", "gid",
+                 "sha1sum", "size",
+                 "blocks", "rdev", "blksize",
+                 ]
+    
+    modemap = {
+        stat.S_IFDIR: 'directory',
+        stat.S_IFREG: 'regular file',
+        stat.S_IFLNK: 'symbolic link',
+        stat.S_IFBLK: 'block device',
+        stat.S_IFCHR: 'character device',
+        stat.S_IFIFO: 'pipe',
+        stat.S_IFSOCK: 'socket',
+        }
                  
     def __init__(self, vi = None, revno = None):
         if vi is None:
-            # the init will be done by hand
+            # the init will be done manually
+            # ugly....
             return
         source = forcejoin(vi.area.root, vi.name)
         self.filename = vi.name
@@ -138,7 +154,9 @@ class RevEntry(object):
         self.revno = revno
         st = os.lstat(source)
 
+        # Read mandatory stat attributes
         self.filetype = stat.S_IFMT(st.st_mode)
+        # Read filecontents
         if self.filetype == stat.S_IFREG:
             self.filecontents = file(source, "r").read()
         elif self.filetype == stat.S_IFDIR:
@@ -147,49 +165,39 @@ class RevEntry(object):
             self.filecontents = os.readlink(source)
         else:
             self.filecontents = ""
+        psha = sha.new(self.filecontents)
+        self.sha1sum = psha.hexdigest()
+        self.inode = st.st_ino
+        self.device = st.st_dev
+        self.nlink = st.st_nlink
         self.mode = st.st_mode
         self.mtime = st.st_mtime
         self.atime = st.st_atime
+        self.ctime = st.st_ctime
         self.uid = st.st_uid
         self.gid = st.st_gid
-        self.rdev = st.st_rdev
-        return
+        self.size = st.st_size
 
-    def _mknod(path, mode, device):
-        if hasattr(os, mknod):
-            of.mknod(path, mode, device)
+        # Try to read optional stat components
+        self.blocks = getattr(st, 'st_blocks', None)
+        self.blksize = getattr(st, 'st_blksize', None)
+        if self.filetype == stat.S_IFBLK or self.filetype == stat.S_IFCHR:
+            self.rdev = getattr(st, 'st_rdev', None)
         else:
-            if mode & stat.S_IFCHR:
-                os.system("mknod '%s' c %u %u" % (path, self.rdev >> 8, self.rdev & 255))
-            elif mode & stat.S_IFBLK:
-                os.system("mknod '%s' b %u %u" % (newname, self.rdev >> 8, self.rdev & 255))
-            else:
-                raise TypeError, "Unknown mode passed to mknod!"
+            self.rdev = None
         return
 
-    _mknod = staticmethod(_mknod)
-
     def _diffdata(older, newer, ofname=None, nfname=None, oftime=None, nftime=None):
-        if hasattr(difflib, 'unified_diff'):
-            a = older.splitlines(True)
-            b = newer.splitlines(True)
-            differ = difflib.unified_diff(
-                a, b,
-                ofname, nfname,
-                oftime, nftime,
-                )
-            data = "".join(differ)
-        else:
-            f1 = file("/tmp/a", "w")
-            f2 = file("/tmp/b", "w")
-            f1.write(older.filecontents)
-            f2.write(self.filecontents)
-            f1.close()
-            f2.close()
-            (status, output) = commands.getstatusoutput("diff -u --label='%s' --label='%s' /tmp/a /tmp/b" % (ofname, nfname))
-            data = output
+        a = older.splitlines(True)
+        b = newer.splitlines(True)
+        differ = difflib.unified_diff(
+            a, b,
+            ofname, nfname,
+            oftime, nftime,
+            )
+        data = "".join(differ)
         return data
-
+    
     _diffdata = staticmethod(_diffdata)
     
     def to_filesys(self, destdir=None, use_dirs=1):
@@ -202,6 +210,7 @@ class RevEntry(object):
 
         """
 
+        self._check_sum()
         if destdir is None:
             target = self.filename
         else:
@@ -212,16 +221,31 @@ class RevEntry(object):
 
         if self.filetype not in (stat.S_IFREG, stat.S_IFLNK, stat.S_IFCHR,
                                  stat.S_IFBLK, stat.S_IFIFO, stat.S_IFDIR):
-            raise TypeError("Can't apply!")
+            print >>sys.stderr, "Skipping: file type %s (%s) cannot be restored." \
+                  % (self.modemap[self.filetype], self.filename)
+            return
+
+        if (self.filetype == stat.S_IFBLK or self.filetype == stat.S_IFCHR) \
+           and self.rdev is None:
+            print >>sys.stderr, "Skipping: st_rdev information not available, can't restore device file!"
+            return
 
         do_create  = True
         do_payload = True
         do_attrs   = True
         do_rename  = True
         
-        if self.filetype == stat.S_IFDIR and os.path.exists(target):
+        # check for parent existence
+        targetbase = os.path.dirname(target)
+        if not rexists(targetbase):
+            try:
+                os.makedirs(targetbase)
+            except OSError, e:
+                print >>sys.stderr, "Warning: While making intermediate directories: %s" % e
+
+        if self.filetype == stat.S_IFDIR and rexists(target):
             if not os.path.islink(target) and not os.path.isdir(target):
-                print >>sys.stderr, "Can't overwrite non-directory with directory!"
+                print >>sys.stderr, "Error: Can't overwrite non-directory with directory!"
                 return
             else: # The directory already exists; we must restore ownership and attrs
                 do_create  = False
@@ -234,11 +258,11 @@ class RevEntry(object):
                 return
             newname = "%s.%07d" % (target, random.randint(0, 999999))
             retries = 0
-            while os.path.exists(newname) and retries < 1000:
+            while rexists(newname) and retries < 1000:
                 newname = "%s.%07d" % (target, random.randint(0, 999999))
                 retries += 1
-            if os.path.exists(newname):
-                print >>sys.stderr, "Can't create a temporary filename! Programmer error or race attack?"
+            if rexists(newname):
+                print >>sys.stderr, "Error: Can't create a temporary filename! Programmer error or race attack?"
                 return
 
         oldumask = os.umask(0777)
@@ -249,7 +273,8 @@ class RevEntry(object):
             # 1. creation of item; can fail; fatal
             # 2. if item is file, write contents; can fail; fatal
             # 3. change ownership; can fail; non-fatal
-            # 4. change permissions and timestamps; shouldn't fail; fatal
+            # 4. change permissions and timestamps; shouldn't fail
+            #    except for symlinks; fatal
             # 5. rename to target; can fail; fatal
 
             # Step 1
@@ -262,14 +287,14 @@ class RevEntry(object):
                     elif self.filetype == stat.S_IFLNK:
                         os.symlink(self.filecontents, newname)
                     elif self.filetype == stat.S_IFCHR or self.filetype == stat.S_IFBLK:
-                        self._mknod(newname, self.mode, self.rdev)
+                        os.mknod(newname, self.mode, self.rdev)
                     elif self.filetype == stat.S_IFDIR:
                         os.mkdir(newname, 0)
                     else:
-                        raise TypeError, "Don't know how to handle file type %s!" % self.filetype
+                        raise TypeError, "Programming error: shouldn't have to handle file type %s!" % self.modemap[self.filetype]
                         return
                 except OSError, e:
-                    print >>sys.stderr, "Abort: error '%s' while creating temporary file." % e
+                    print >>sys.stderr, "Error: error '%s' while creating temporary file." % e
                     return
                 else:
                     must_remove_new = True
@@ -281,7 +306,7 @@ class RevEntry(object):
                         os.write(fd, self.filecontents)
                         os.close(fd)
                 except OSError, e:
-                    print >>sys.stderr, "Abort: error '%s' while writing file contents." % e
+                    print >>sys.stderr, "Error: error '%s' while writing file contents." % e
                     return
             # Step 3
             if do_attrs:
@@ -289,19 +314,22 @@ class RevEntry(object):
                     os.lchown(newname, self.uid, self.gid)
                 except OSError, e:
                     print >>sys.stderr, "Warning: error '%s' while modifying temporary file ownership." % e
-                # Step 4
-                try:
-                    os.chmod(newname, self.mode)
-                    os.utime(newname, (self.atime, self.mtime))
-                except OSError, e:
-                    print >>sys.stderr, "Abort: error '%s' while modifying temporary file attributes." % e
-                    return
+
+                # WARNING: don't chmod for symlinks, it acts on the target!!!
+                if not self.filetype == stat.S_IFLNK:
+                    # Step 4
+                    try:
+                        os.chmod(newname, self.mode)
+                        os.utime(newname, (self.atime, self.mtime))
+                    except OSError, e:
+                        print >>sys.stderr, "Error: error '%s' while modifying temporary file attributes." % e
+                        return
             # Step 5
             if do_rename:
                 try:
                     os.rename(newname, target)
                 except OSError, e:
-                    print >>sys.stderr, "Abort: error '%s' while renaming" % e
+                    print >>sys.stderr, "Error: error '%s' while renaming" % e
                     raise
                 else:
                     must_remove_new = False # We managed to finish!
@@ -311,10 +339,10 @@ class RevEntry(object):
                 try:
                     os.unlink(newname)
                 except OSError, e:
-                    print >>sys.stderr, "Error while cleaning-up: %s" % e
+                    print >>sys.stderr, "Error: while cleaning-up: %s" % e
         return
         
-    def diff(self, older):
+    def diff(self, older, options=None):
         obuff = StringIO()
         if not isinstance(older, RevEntry):
             raise TypeError("Invalid diff!")
@@ -326,25 +354,30 @@ class RevEntry(object):
             newrev = "rev %s" % self.revno
         orev = "rev %s" % older.revno
         if self.filetype != older.filetype:
-            return "File type has changed: from %s to %s" % (older.filetype, self.filetype)
+            return "File type has changed: from %s to %s" % \
+                   (older.modemap[older.filetype],
+                    self.modemap[self.filetype])
         if self.filetype == stat.S_IFREG or self.filetype == stat.S_IFDIR:
             if self.filecontents != older.filecontents:
-                obuff.write("== File contents diff: ==\n")
-                data = self._diffdata(
-                    older.filecontents, self.filecontents,
-                    older.filename, self.filename,
-                    "%s (%s)" % (time.ctime(older.mtime), orev),
-                    "%s (%s)" % (time.ctime(self.mtime), newrev)
-                    )
-                obuff.write(data)
-                obuff.write("\n")
+                if self.printablepayload() and older.printablepayload():
+                    obuff.write("== File contents diff: ==\n")
+                    data = self._diffdata(
+                        older.filecontents, self.filecontents,
+                        older.filename, self.filename,
+                        "%s (%s)" % (time.ctime(older.mtime), orev),
+                        "%s (%s)" % (time.ctime(self.mtime), newrev)
+                        )
+                    obuff.write(data)
+                    obuff.write("\n")
+                else:
+                    obuff.write("Binary files %s and %s differ\n" % (older.filename, self.filename))
         elif self.filetype == stat.S_IFLNK:
             if self.filecontents != older.filecontents:
                 obuff.write("Symlink target changed from '%s' to '%s'" % (older.filecontents, self.filecontents))
         else:
             obuff.write("Don't know how to diff")
         obuff.write("== File metadata diff ==\n")
-        for i in ('mtime', 'atime', 'uid', 'gid', 'mode'):
+        for i in ('ctime', 'mtime', 'uid', 'gid', 'mode'):
             oval = getattr(older, i)
             nval = getattr(self, i)
             if oval != nval:
@@ -357,12 +390,18 @@ class RevEntry(object):
             return NotImplemented
         # Do not test atime, it's irrelevant
         for i in ('filename', 'filetype', 'mode',
-                  'mtime', 'uid', 'gid', 'rdev',
-                  'filecontents'):
+                  'mtime', 'uid', 'gid',
+                  'filecontents', 'rdev'):
             if getattr(self, i) != getattr(other, i):
                 return False
         return True
 
+    def printablepayload(self):
+        for i in self.filecontents:
+            if i not in string.printable:
+                return False
+        return True
+    
     def isdir(self):
         return self.filetype == stat.S_IFDIR
 
@@ -417,17 +456,61 @@ class RevEntry(object):
         tchar += mapbit(self.mode, stat.S_IWOTH, 'w')
         tchar += mapbit(self.mode, stat.S_IXOTH, 'x')
         return tchar
+
+    def _check_sum(self):
+        psha = sha.new(self.filecontents)
+        psum = psha.hexdigest()
+        if psum != self.sha1sum:
+            raise ValueError, "Invalid checksum in file contents (%s != %s)!" % (psum, self.sha1sum)
+        return
+
+    def stat(self):
+        obuf=StringIO()
+        obuf.write("  File: `%s'\n" % self.filename)
+        obuf.write("  Size: %-15d Blocks: " % self.size)
+        if self.blocks is None:
+            obuf.write("%-10s" % "N/A")
+        else:
+            obuf.write("%-10d" % self.blocks)
+        obuf.write("IO Block: ")
+        if self.blksize is None:
+            obuf.write("%-6s" % "N/A")
+        else:
+            obuf.write("%-6d " % self.blksize)
+        obuf.write(self.modemap[self.filetype])
+        obuf.write("\n")
+        obuf.write("Device: %3xh/%3dd       Inode: %-11d Links: %d\n" %
+                   (self.device, self.device, self.inode, self.nlink))
+        obuf.write("Access: (%04o/%s) Uid: (%5d/%8s)   Gid: (%5d/%8s)\n" %
+                   (stat.S_IMODE(self.mode), self.mode2str(),
+                    self.uid, "",
+                    self.gid, "")
+                   )
+        for i in (('Access', self.atime),
+                  ('Modify', self.mtime),
+                  ('Change', self.ctime)):
+            ovar = DateTime.localtime(i[1])
+            obuf.write("%s: %s %s\n" % (i[0], ovar.strftime("%Y-%m-%d %T.000000000"), ovar.tz))
+        return obuf.getvalue()
             
 class AreaRev(object):
     __slots__ = ["area", "revno", "logmsg", "uid", "gid",
-                 "commiter", "_ctime", "itemids"]
+                 "commiter", "_ctime", "itemids", "server"]
 
-    def __init__(self, area=None, logmsg=None, commiter=None):
+    def __init__(self, area=None, logmsg=None, commiter=None,
+                 server=None):
         if area is None and logmsg is None:
             # manual initialization
             return
-        self.area = area.id
-        self.revno = area.revno + 1
+        self.area = area.name
+        if area is None or area.revno is None:
+            self.revno = 1
+        else:
+            self.revno = area.revno + 1
+        if server is None:
+            self.server = os.uname()[1]
+        else:
+            self.server = server
         self.logmsg = logmsg
         self._ctime = DateTime.utc()
         self.uid = os.getuid()
@@ -435,7 +518,7 @@ class AreaRev(object):
         if commiter is None:
             try:
                 self.commiter = os.getlogin()
-            except OSError, e:
+            except OSError:
                 self.commiter = '<unknown>'
         else:
             self.commiter = commiter
